"""Slices Basic Infrastructure Machine Info API."""

from collections.abc import Collection
from datetime import datetime
from typing import Annotated

from httpx import Client, HTTPError
from msgspec import Meta, Struct
from msgspec.json import Decoder as JsonDecoder

from .exceptions import from_httpx_exception
from .pagination import PagedCollection


class GeoLocation(Struct, tag="geo"):
    """A location on planet earth in latitude/longitude coordinates.

    For convenience, the country is provided as well.
    """

    country: str
    """Country code in ISO 3166-1 alpha-2."""
    lat: float
    long: float


class GridLocation(Struct, tag="grid"):
    """A relative location on a grid."""

    x: float
    y: float
    z: float


class RackAndPosition(Struct, tag="rack+pos"):
    """The location of a machine, by specifying rack ID and position within the rack."""

    rack: str
    """ID of the rack."""
    position: int
    """Position of a machine within the rack."""


AnyLocation = GeoLocation | GridLocation | RackAndPosition


class Availability(Struct):
    """Info about availability.

    Is it in use or free?
    """

    available: bool
    """Is available?"""
    updated_at: Annotated[datetime, Meta(tz=True)]
    """Timestamp when info about availability was last updated"""


class Machine(Struct):
    """Info about physical Machine."""

    id: str
    """TypeID uniquely identifying a machine"""

    name: Annotated[
        str,
        Meta(
            min_length=2,
            max_length=30,
            pattern="[a-zA-Z0-9][a-zA-Z0-9._-]+",
        ),
    ]
    """Name of the machine. Typically an infrastructure specific ID."""

    description: str | None
    """Description of this machine"""

    available: Availability
    """Info about availability of this machine"""

    tags: list[str] = []
    flavors: list[str] = []
    """Flavors of this machine."""
    location: list[AnyLocation] = []
    """Location of this machine."""


_machine_response_decoder = JsonDecoder(Machine)
_machines_response_decoder = JsonDecoder(list[Machine])


class BiMachineClient:
    """Basic Infrastructure Machine Info API Client."""

    API_BASE_PATH = "/apis/bi.slices.eu/v1alpha2/"

    def __init__(self, client: Client) -> None:
        """Init Basic Infrastructure Machine Info Client."""
        self._client = client
        self._base_path = f"{self.API_BASE_PATH}machines/"
        assert client.base_url is not None, "Expected base_url to be set"

    def get(self, id: str) -> Machine:
        """Get machine info by ID."""
        response = self._client.get(f"{self._base_path}{id}")

        try:
            response.raise_for_status()
            return _machine_response_decoder.decode(response.content)
        except HTTPError as err:
            raise from_httpx_exception(err) from err

    def list(self) -> Collection[Machine]:
        """List machines."""

        params = {}

        return PagedCollection("machine", Machine, self._base_path, self._client, params)
