"""Slices Basic Infrastructure Resource API."""

from collections.abc import Collection
from datetime import datetime
from enum import Enum
from typing import Annotated, Any, Literal, Optional

from httpx import Client, HTTPError
from msgspec import Meta, Struct, field
from msgspec.json import Decoder as JsonDecoder

from .bi_task import TaskStatus, _task_status_decoder, _tasks_status_decoder
from .exceptions import from_httpx_exception
from .pagination import PagedCollection
from .project import ProjectClient
from .util.fabric import connect as connect_with_fabric
from .utils import json_encoder


class ResourceState(str, Enum):
    """Resource state."""

    """
    allowed enum values
    """
    UNKNOWN = "unknown"
    INACTIVE = "inactive"
    RESERVED = "reserved"
    DELETING = "deleting"
    DELETED = "deleted"
    FAILED = "failed"
    IMAGING = "imaging"
    BOOTING = "booting"
    INITIALIZING = "initializing"
    STARTING = "starting"
    UP = "up"
    POWEROFF = "poweroff"
    REBOOTING = "rebooting"


class SSHInfo(Struct):
    """Info about SSH Login to a resource."""

    host: str
    port: int
    username: str

    host_keys: list[str] | None
    """The public key(s) of the target SSH server, in openssh format.

    This list can be empty. It is recommended that it is not.
    The typical line you need to add to ~/.ssh/known_hosts is the host, a space, and then one of these keys.
    """

    description: str | None = None
    jump_proxy: Optional["SSHInfo"] = None

    def connect(self):
        """Create a Fabric connection."""

        return connect_with_fabric(self)


class NetworkInterface(Struct):
    """A network interface describes a connection of a resource to a network."""

    ip: str
    network_id: str
    "ID of the network to which this interface is connected"
    port_id: str
    "Name of the NIC to which this interface is coupled"
    version: int


class ComputeResource(Struct):
    """Compute Resource Base."""

    cluster_id: str
    created_at: Annotated[datetime, Meta(tz=True)]
    "Timestamp when this compute resource was created"

    disk_image_id: str
    "The ID of the image to use for this server instance"

    experiment_id: str
    "TypeID of an Experiment where this resource is a part of"

    expires_at: Annotated[datetime, Meta(tz=True)] | None
    flavor_id: str
    "The flavor reference"

    friendly_name: str
    """User readable name, must conform to the following rules:
    - Can only contain alphanumeric characters (a-z 0-9) or a minus sign (-)
    - Must be at least 2 characters long, cannot be longer than 63 characters
    - Must not start or end with a minus sign (-)
    """

    id: str
    "TypeID uniquely identifying a resource"

    network_interfaces: list[NetworkInterface]
    private_ipv4: str | None
    private_ipv6: str | None
    public_ipv4: str | None
    public_ipv6: str | None
    ssh_authorized_keys: list[str] | None
    "Public SSH Keys that can login to the resource."

    ssh_logins: list[SSHInfo]
    "SSH Logins for this resource."
    status: ResourceState
    terminated_at: Annotated[datetime, Meta(tz=True)] | None
    userdata: str | None = None
    console_url: str | None = None


_compute_resource_decoder = JsonDecoder(ComputeResource)


# class Task(Struct):
#     """Task status of a long running background task."""

#     id: str
#     status: str
#     total: float | int | None = None
#     completed: float | int | None = None
#     description: str | None = None
#     result: str | None = None


# _task_decoder = JsonDecoder(Task)


class Route(Struct):
    """A route describes a route to a network."""

    to: str
    via: str
    metric: Annotated[int, Meta(gt=0)] | None = None
    mtu: int | None = None
    scope: str | None = None
    table: int | None = None


class NameServers(Struct):
    """Name Servers."""

    addresses: list[str]
    search: list[str] | None = None


class NetworkInterfaceCreate(Struct):
    """A network interface describes a connection of a resource to a network."""

    network_id: str
    port_id: str

    addresses: list[str] | None = None
    dhcp4: bool | None = None
    dhcp6: bool | None = None
    nameservers: NameServers | None = None
    routes: list[Route] | None = None


class ComputeResourceCreate(Struct):
    """Request for a new Compute Resource."""

    friendly_name: str
    disk_image: str | None = None
    disk_image_id: str | None = None
    flavor: str | None = None
    flavor_id: str | None = None

    network_interfaces: list[NetworkInterfaceCreate] | None = None
    infra_id: str | None = None
    userdata: str | None = None
    machine_id: str | None = None


class LinkImpairment(Struct):
    """Specification of an impairment.

    Specification of impairment on a single direction of a 2 way link
    (both source and destination), or specification of impairment on a LAN,
    symmetrically impairing traffic from and to each of the nodes (no source and no destination).
    """

    capacity_kbit_per_s: int | None = None
    destination: str | None = None
    latency_ms: int | None = 0
    packet_loss: float | int | None = 0
    source: str | None = None


class BareMetalLinkCreate(Struct):
    """A link between two or more resources."""

    friendly_name: str
    network_id: str

    # cross infra link info
    vlan_tag: int | None = None
    cross_infra_link_capacity_mbit_per_s: int | None = None

    # bare metal
    best_effort: bool | None = False
    impairment: list[LinkImpairment] | None = None
    interswitch: bool | None = False
    link_multiplexing: bool | None = False
    no_mac_learning: bool | None = False
    share_lan_name: str | None = None
    type: str | None = "lan"
    vlan_tagging: bool | None = False


class Token(Struct):
    """A token."""

    jwt: str


class NewResourcesRequest(Struct):
    """Request for new resources in a given experiment."""

    expires_at: datetime
    resources: list[ComputeResourceCreate]
    ssh_authorized_keys: list[str]
    links: list[BareMetalLinkCreate] = field(default_factory=list)
    tokens: list[Token] = field(default_factory=list)


class NewComputeResource(ComputeResource):
    """Newly created resource."""

    task_id: str | None = None


class NewResourcesResponse(Struct):
    """Newly created resources response."""

    request_id: str
    resources: list[NewComputeResource]


_new_resources_response_decoder = JsonDecoder(NewResourcesResponse)


class Snapshot(Struct):
    """Resource Snapshot."""

    name: str
    description: str | None = None
    created: Annotated[datetime, Meta(tz=True)] | None = None
    includes_ram: bool | None = None


_snapshots_response_decoder = JsonDecoder(list[Snapshot])


class ConsoleLoginDetails(Struct):
    """Console login details."""

    username: str
    password: str


class BiResourceClient:
    """Basic Infrastructure Resources API Client."""

    API_BASE_PATH = "/apis/bi.slices.eu/v1alpha2/"

    def __init__(self, client: Client, experiment_id: str, project_id: str, project_client: ProjectClient) -> None:
        """Create Basic Infrastructure Resources Client."""
        self._client = client

        self._experiment_id = experiment_id
        self._project_id = project_id
        self._project_client = project_client

        self._base_path = f"{self.API_BASE_PATH}projects/{project_id}/experiments/{experiment_id}/resources/"

        assert client.base_url is not None, "Expected base_url to be set"

    def get(self, id: str) -> ComputeResource:
        """Get resource by ID."""
        response = self._client.get(f"{self._base_path}{id}")

        try:
            response.raise_for_status()
            return _compute_resource_decoder.decode(response.content)
        except HTTPError as err:
            raise from_httpx_exception(err) from err

    def update_all(self, new_expires_at: datetime) -> list[TaskStatus]:
        """Update resource by ID."""
        response = self._client.patch(
            f"{self._base_path}",
            json={"expires_at": new_expires_at.isoformat()},
        )

        try:
            response.raise_for_status()
            return _tasks_status_decoder.decode(response.content)
        except HTTPError as err:
            raise from_httpx_exception(err) from err

    def update(self, id: str, new_expires_at: datetime) -> TaskStatus:
        """Update resource by ID."""
        response = self._client.patch(
            f"{self._base_path}{id}",
            json={"expires_at": new_expires_at.isoformat()},
        )

        try:
            response.raise_for_status()
            return _task_status_decoder.decode(response.content)
        except HTTPError as err:
            raise from_httpx_exception(err) from err

    def get_console_login_details(self, id: str) -> ConsoleLoginDetails:
        """Get resource by ID."""
        response = self._client.get(f"{self._base_path}{id}/console/login_details")

        try:
            response.raise_for_status()
        except HTTPError as err:
            raise from_httpx_exception(err) from err

        return JsonDecoder(ConsoleLoginDetails).decode(response.content)

    def delete(self, compute_resource: ComputeResource) -> TaskStatus:
        """Delete compute resource."""
        return self.delete_by_id(compute_resource.id)

    def delete_by_id(self, id: str) -> TaskStatus:
        """Delete compute resource."""
        response = self._client.delete(f"{self._base_path}{id}")

        try:
            response.raise_for_status()
            return _task_status_decoder.decode(response.content)
        except HTTPError as err:
            raise from_httpx_exception(err) from err

    def list(
        self,
        include_expired: bool | None = None,
        include_terminated: bool | None = None,
    ) -> Collection[ComputeResource]:
        """List resources."""

        params = {}
        if include_expired is not None:
            params["include_expired"] = "true" if include_expired else "false"
        if include_terminated is not None:
            params["include_terminated"] = "true" if include_terminated else "false"

        return PagedCollection("resource", ComputeResource, self._base_path, self._client, params)

    def reboot(self, resource_id: str, type: Literal["SOFT", "HARD"]):
        """Reboot the resource."""
        return self._perform_action(resource_id, {"reboot": {"type": type}})

    def create_snapshot(self, resource_id: str, snapshot_name: str, snapshot_description: str | None = None):
        """Create a snapshot."""
        details = {"snapshot_name": snapshot_name}
        if snapshot_description:
            details["snapshot_description"] = snapshot_description
        return self._perform_action(resource_id, {"createSnapshot": details})

    def revert_to_snapshot(self, resource_id: str, snapshot_name: str):
        """Revert to a snapshot."""
        return self._perform_action(resource_id, {"revertToSnapshot": {"snapshot_name": snapshot_name}})

    def list_snapshots(self, resource_id: str) -> Collection[Snapshot]:
        """List snapshots."""
        response = self._client.get(f"{self._base_path}{resource_id}/snapshots")

        try:
            response.raise_for_status()
        except HTTPError as err:
            raise from_httpx_exception(err) from err

        return _snapshots_response_decoder.decode(response.content)

    def _perform_action(self, resource_id: str, json: Any) -> TaskStatus:
        """Perform action on resource."""
        response = self._client.post(f"{self._base_path}{resource_id}/action", json=json)

        try:
            response.raise_for_status()
            return _task_status_decoder.decode(response.content)
        except HTTPError as err:
            raise from_httpx_exception(err) from err

    def create(self, new_resources_request: NewResourcesRequest) -> NewResourcesResponse:
        """Create new resources."""

        response = self._client.post(
            self._base_path,
            headers={"Content-Type": "application/json"},
            content=json_encoder.encode(new_resources_request),
        )

        try:
            response.raise_for_status()
        except HTTPError as err:
            raise from_httpx_exception(err) from err

        return _new_resources_response_decoder.decode(response.content)
