"""Slices BI Task API."""

from httpx import Client, HTTPError
from msgspec import Struct
from msgspec.json import Decoder as JsonDecoder

from .exceptions import from_httpx_exception


class TaskStatus(Struct):
    """Task status of a long-running background task."""

    id: str
    status: str

    completed: float | int | None = None
    total: float | int | None = None
    description: str | None = None
    result: str | None = None


class TaskLogLine(Struct):
    """Log line."""

    n: int  # Line number
    t: str  # Line text


_task_status_decoder = JsonDecoder(TaskStatus)
_tasks_status_decoder = JsonDecoder(list[TaskStatus])
_task_log_lines_decoder = JsonDecoder(list[TaskLogLine])


class BiTaskClient:
    """Basic Infrastructure Task API Client."""

    API_BASE_PATH = "/apis/bi.slices.eu/v1alpha2/tasks/"

    def __init__(self, client: Client):
        """Create Task Client."""
        self._client = client

    def get_task(self, task_id: str) -> TaskStatus:
        """Get the status of a task."""
        response = self._client.get(f"{self.API_BASE_PATH}{task_id}")
        try:
            response.raise_for_status()
        except HTTPError as err:
            raise from_httpx_exception(err) from err
        return _task_status_decoder.decode(response.content)

    def get_tasks(self, task_ids: list[str]) -> list[TaskStatus]:
        """Get the status of multiple tasks."""
        response = self._client.get(
            f"{self.API_BASE_PATH}",
            params={"task_ids": task_ids},
        )
        try:
            response.raise_for_status()
        except HTTPError as err:
            raise from_httpx_exception(err) from err
        return _tasks_status_decoder.decode(response.content)

    def get_task_logs(
        self,
        task_id: str,
        start: int | None = None,
        limit: int | None = None,
    ) -> list[TaskLogLine]:
        """Get the log of a task."""
        params = {}
        if start:
            params["start"] = start
        if limit:
            params["limit"] = limit

        response = self._client.get(f"{self.API_BASE_PATH}{task_id}/logs", params=params)
        try:
            response.raise_for_status()
        except HTTPError as err:
            raise from_httpx_exception(err) from err

        return _task_log_lines_decoder.decode(response.content)
