"""Authentication flow."""

import contextlib
import socket
import string
import time
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime, timedelta
from secrets import SystemRandom

import msgspec
from authlib.integrations.httpx_client import OAuth2Client, OAuthError
from authlib.oauth2.rfc8628 import DEVICE_CODE_GRANT_TYPE

from .config import settings
from .server import wait_for_authorization_response
from .token import OAuthToken

CODE_VERIFIER_CHARS = string.ascii_letters + string.digits + "-._~"
CODE_VERIFIER_RANDOM = SystemRandom()

DEVICE_FLOW_POLLING_FREQUENCY = 2


oauth_client = OAuth2Client(
    client_id=settings.client_id,
    client_secret=settings.client_secret,
    scope=settings.scope,
    token_endpoint_auth_method="client_secret_post",  # noqa: S106
)


def _generate_code_verifier():
    return "".join(CODE_VERIFIER_RANDOM.choice(CODE_VERIFIER_CHARS) for _ in range(128))


class AuthenticationFlow:
    """Authentication Flow."""

    authorization_url: str
    """URL that the user must visit to authenticate."""

    def __init__(self) -> None:
        """Initialize the AuthenticationFlow."""

        callback_port = self._find_open_port()

        self._callback_server_executor = ThreadPoolExecutor(max_workers=1, thread_name_prefix="AuthFlowServer")
        self._callback_server_future = self._callback_server_executor.submit(
            wait_for_authorization_response, "localhost", callback_port
        )

        code_verifier = _generate_code_verifier()

        self._redirect_uri = f"http://localhost:{callback_port}/callback"
        self.authorization_url, self._state = oauth_client.create_authorization_url(
            settings.authorization_url,
            redirect_uri=self._redirect_uri,
            code_verifier=code_verifier,
        )

    def wait_for_authorization_token(self) -> OAuthToken:
        """Blocking method which waits for the user to authenticate and returns the token."""
        authorization_response = self._callback_server_future.result()

        token = oauth_client.fetch_token(
            settings.token_endpoint_url,
            state=self._state,
            authorization_response=authorization_response,
            redirect_uri=self._redirect_uri,
        )
        if token:
            return msgspec.convert(token, OAuthToken)

        raise ValueError("Did not receive a token")

    @staticmethod
    def _find_open_port(start_port=None, stop_port=None):
        start = start_port or 8080
        stop = stop_port or start + 100
        for port in range(start, stop):
            with contextlib.closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
                try:
                    sock.bind(("127.0.0.1", port))
                    sock.listen(1)
                except OSError:
                    is_open = False
                else:
                    is_open = True
            if is_open:
                return port
        raise ConnectionError("Could not find an open port")


class DeviceFlow:
    """Device Authentication Flow."""

    def __init__(self):
        """Perform OAuth Login flow and return authorization code."""

        code_verifier = _generate_code_verifier()

        _, self._state = oauth_client.create_authorization_url(
            settings.authorization_url,
            code_verifier=code_verifier,
            grant_type=DEVICE_CODE_GRANT_TYPE,
        )

        device_code_response = oauth_client.request(
            "POST",
            settings.device_authorization_url,
            withhold_token=True,
            data={
                "client_id": settings.client_id,
                "client_secret": settings.client_secret,
                "scope": settings.scope,
            },
        )

        self._device_code_info = device_code_response.json()

        self.authorization_url = self._device_code_info["verification_uri_complete"]
        self._expires_at = datetime.now() + timedelta(seconds=self._device_code_info["expires_in"])

    def wait_for_authorization_token(self) -> OAuthToken:
        """Blocking method which waits for the user to authenticate and returns the token."""

        while self._expires_at > datetime.now():
            time.sleep(DEVICE_FLOW_POLLING_FREQUENCY)
            try:
                token = oauth_client.fetch_token(
                    settings.token_endpoint_url,
                    state=self._state,
                    grant_type=DEVICE_CODE_GRANT_TYPE,
                    device_code=self._device_code_info["device_code"],
                )
                if token:
                    return msgspec.convert(token, OAuthToken)
            except OAuthError as err:
                if err.error != "authorization_pending":
                    raise ValueError("Exception while waiting for device authorization") from err

        raise TimeoutError("Device login flow timed out.")


class RefreshFlow:
    """Refresh Flow."""

    def perform_refresh(self, refresh_token: str):
        """Perform OAuth Login flow and return authorization code."""

        return msgspec.convert(oauth_client.refresh_token(settings.token_endpoint_url, refresh_token), OAuthToken)
